import click
import os
import configparser
from gesture_app import GESTURE_APP_HOME

config_file = os.path.join(GESTURE_APP_HOME, "cfg")
config = configparser.ConfigParser()


def checkIfUserValid(name):
    if (name is None or name == ""):
        return True
    return False


def addConfig(key, value):
    """Set global configuration for this app."""
    config.read(config_file)
    section, key = key.split(".")
    if not config.has_section(section):
        config.add_section(section)
    config.set(section, key, value)
    with open(config_file, "w") as configfile:
        config.write(configfile)


def getConfig(key):
    """Get global configuration for this app."""
    config.read(config_file)
    section, key = key.split(".")
    if not config.has_section(section):
        return None
    try:
        return config.get(section, key)
    except KeyError:
        return None
    except configparser.NoOptionError:
        return None


def printSectionConfig(section_name):
    section_name = section_name[:-1]
    config.read(config_file)
    try:
        for name, value in config.items(section_name):
            click.echo('  %s = %s' % (name, value))
    except configparser.NoSectionError:
        click.echo('No such configuration exists.')


import cv2
import time
from gesture_app.model.clean_image import CleanImage
import copy
import numpy as np
import keyboard
import os
import tensorflow as tf
import importlib.resources as pkg_resources
import warnings
warnings.filterwarnings("ignore", category=FutureWarning)


class GesturePrediction:

    @staticmethod
    def file_path(relative_path):
        __dir = os.path.dirname(os.path.abspath(__file__))
        split_path = relative_path.split("/")
        new_path = os.path.join(__dir, *split_path)
        return new_path

    def __init__(self, keyboard_command):
        self.interpreter = tf.lite.Interpreter(model_path=self.file_path("tf_lite_model.tflite"))
        self.interpreter.allocate_tensors()

        # Get input and output tensors.
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()

        print("Loaded model from disk")
        self.gestures = {'high_five': 1, 'null': 0}
        self.inv_gestures = {v: k for k, v in self.gestures.items()}
        self.image_cleaner = CleanImage()
        self.keyboard_command = keyboard_command

    def live_video(self):
        t1 = time.time()
        video_capture = cv2.VideoCapture(0)
        capture_background_flag = True
        count = 0
        while True:
            t2 = time.time()
            ret, self.frame = video_capture.read()
            if t2 - t1 > 5:
                capture_background_flag = False
            if cv2.waitKey(1) & 0xFF == ord('r'):
                t1 = time.time()
                capture_background_flag = True
            if capture_background_flag:
                self.image_cleaner.subtract_background = self.image_cleaner.refresh_background()
                continue
            self.image_cleaner.frame = self.frame
            self.image_cleaner.process()
            self.frame = self.image_cleaner.frame
            prediction = self.get_prediction()
            if prediction[0] == 'high_five':
                count += 1
                if count > 2:
                    keyboard.press_and_release(self.keyboard_command)
                    time.sleep(1)
                    count = 0
            else:
                count = 0
            cv2.putText(self.frame, f"Press Q to quit | Press R to refresh", (0, 20),
                        self.image_cleaner.font, self.image_cleaner.fontScale, (255, 255, 255))
            cv2.putText(self.frame, f"Prediction: {prediction}", (0, 40),
                        self.image_cleaner.font, self.image_cleaner.fontScale, (255, 255, 255))
            cv2.imshow('frame', self.frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        video_capture.release()
        cv2.destroyAllWindows()

    def get_prediction(self):
        roi = self.frame
        roi = cv2.resize(roi, (224, 224))
        roi = np.expand_dims(roi, axis=0)
        roi_copy = copy.deepcopy(roi)
        roi = np.stack((roi, roi_copy, roi_copy), axis=3).astype(np.float32)
        self.interpreter.set_tensor(self.input_details[0]['index'], roi)
        self.interpreter.invoke()
        prediction = self.interpreter.get_tensor(self.output_details[0]['index'])
        prediction_num = prediction[0][0]
        if prediction_num > 0.7:
            prediction = self.inv_gestures[1]
        else:
            prediction = self.inv_gestures[0]
        return prediction, 100 * prediction_num



import cv2
import numpy as np
import time
import copy

class CleanImage:
    def __init__(self):
        self.cap_region_x_begin = 0.5  # start point/total width
        self.cap_region_y_end = 0.8  # start point/total width
        self.threshold = 90  # binary threshold
        self.blurValue = 31  # GaussianBlur parameter
        self.bgSubThreshold = 50
        self.image_size = (224, 224)
        self.learningRate = 0.01
        self.font = cv2.FONT_HERSHEY_SIMPLEX
        self.fontScale = 1
        self.thickness = 2
        self.subtract_background = self.refresh_background()
        self.kernel = np.ones((2, 2), np.uint8)
        self.lower_skin = np.array([0, 36, 0], dtype=np.uint8)
        self.upper_skin = np.array([57, 158, 255], dtype=np.uint8)

    def live_video(self):
        t1 = time.time()
        video_capture = cv2.VideoCapture(0)
        self.video_framerate = video_capture.get(cv2.CAP_PROP_FPS)
        capture_background_flag = True
        while True:
            t2 = time.time()
            ret, self.frame = video_capture.read()
            if t2 - t1 > 2:
                capture_background_flag = False
            if cv2.waitKey(1) & 0xFF == ord('r'):
                t1 = time.time()
                capture_background_flag = True
            if capture_background_flag:
                self.subtract_background = self.refresh_background()
                continue
            self.process()
            cv2.imshow('frame', self.frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
        video_capture.release()
        cv2.destroyAllWindows()

    def process(self):
        frame = cv2.bilateralFilter(self.frame, 5, 50, 100)
        frame = cv2.flip(frame, 1)
        img = frame[0:int(self.cap_region_y_end * frame.shape[0]),
                    int(self.cap_region_x_begin * frame.shape[1]):frame.shape[1]]
        img_ = cv2.GaussianBlur(img, (self.blurValue, self.blurValue), 0)
        img = cv2.dilate(img_, self.kernel, iterations=4)
        mask = self.subtract_background.apply(img, learningRate=self.learningRate)
        img_copy = copy.deepcopy(img_)
        img = mask
        self.frame = img
        # self.find_contours(img_copy)
        # cv2.putText(self.frame, f"'Q' quit,'R' refresh FPS: {int(self.video_framerate)}", (0, 22),
        #             self.font, self.fontScale, (255, 255, 255))

    def filter_by_skin_color(self, frame):
        hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
        mask = cv2.inRange(hsv, self.lower_skin, self.upper_skin)
        return mask

    def find_contours(self, img_copy):
        contours, _ = cv2.findContours(self.frame, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
        if len(contours) == 0:
            return None
        biggest_contour = max(contours, key=cv2.contourArea)
        biggest_hull = cv2.convexHull(biggest_contour)
        mask = np.zeros(self.frame.shape, dtype='uint8')
        cv2.fillPoly(mask, pts=[biggest_contour], color=(255, 255, 255))
        mask2 = cv2.bitwise_and(self.frame, self.frame, mask=mask)
        self.frame = mask2
        # cv2.drawContours(self.frame, [biggest_contour], -1, (255, 255, 255), 3)
        # cv2.drawContours(self.frame, [biggest_hull], -1, (255, 255, 255), 3)

    @staticmethod
    def refresh_background():
        time.sleep(1)
        return cv2.createBackgroundSubtractorMOG2(history=30, varThreshold=16, detectShadows=False)


# if __name__ == '__main__':
#     image_cleaner = CleanImage()
#     image_cleaner.live_video()



# -*- coding: utf-8 -*-
import os
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, "gesture_app", "__version__.py"), "r") as f:
    exec(f.read(), about)

with open("README.md", "r") as f:
    readme = f.read()

requires = ["Click==7.0", "opencv-python==4.1.0.25", "tensorflow==2.0.0", "keyboard==0.13.5", "numpy==1.18.4"]


def setup_package():
    metadata = dict(
        name=about["__title__"],
        version=about["__version__"],
        description=about["__description__"],
        long_description=readme,
        long_description_content_type="text/markdown",
        url=about["__url__"],
        author=about["__author__"],
        author_email=about["__author_email__"],
        license=about["__license__"],
        packages=find_packages(exclude=("tests",)),
        install_requires=requires,
        entry_points={"console_scripts": ["gesture-app = gesture_app.cli:cli"]},
        classifiers=[
            # Trove classifiers
            # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        include_package_data=True,
    )

    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup

    setup(**metadata)

if __name__ == "__main__":
    setup_package()


import os
import time
import random
import json
import click
from gesture_app import GESTURE_APP_HOME
from gesture_app._utils import checkIfUserValid, addConfig, getConfig, printSectionConfig
from gesture_app.model.gesture_prediction import GesturePrediction

# TODO: handle relative paths. Here the assumption is that the code is being run from the root directory of app
# Solution: https://click.palletsprojects.com/en/7.x/utils/#finding-application-folders (need to verify on other systems)
USER_PREFIX = "user."
ACTION_PREFIX = "action."
GESTURE_PREFIX = "gesture."
MAP_PREFIX = "map."


def build_settings_option_class(settings_instance):
    def set_default(default_name):
        class Cls(click.Option):
            def __init__(self, *args, **kwargs):
                kwargs['default'] = getattr(settings_instance, default_name)
                super(Cls, self).__init__(*args, **kwargs)

            def handle_parse_result(self, ctx, opts, args):
                obj = ctx.find_object(type(settings_instance))
                if obj is None:
                    ctx.obj = settings_instance
                return super(Cls, self).handle_parse_result(ctx, opts, args)

        return Cls

    return set_default


# click.pass_context is a way of using the Click context to pass parameters between commands, like username in our case
class Setting(object):
    def __init__(self):
        self.verbose = False
        self.name = "User"
        self.hsv_thresholds = None
        # passing default actions and gestures
        self.gestures = ['high_five']
        self.actions = [{'actionName': 'screenshot', 'action': 'cmd+shift+3'}]


settings = Setting()
settings_option_cls = build_settings_option_class(settings)
pass_setting = click.make_pass_decorator(Setting, ensure=True)


@click.group("gesture-app")
@click.version_option("0.1.0")
@click.option('--verbose', is_flag=True)
@pass_setting
def cli(ctx, verbose, *args, **kwargs):
    """the app to make your life easy by identifying hand getsures"""
    if verbose:
        click.echo("We are in verbose mode")
    ctx.name = getConfig(USER_PREFIX + "name")
    # print("Context values:" + " " + str(ctx.sname) + " "+ str(ctx.password))


# settings_option_cls can be used here to set any default values using the Settting class, in which case the user can simply press enter
# Ref: https://stackoverflow.com/questions/49511933/better-usage-of-make-pass-decorator-in-python-click
@cli.command()
@click.option('-u', '--name', cls=settings_option_cls('name'), prompt='Please enter your name', help='App user.',
              type=str)
@pass_setting
def register(ctx, name):
    """Register personalised user data"""
    click.echo('Hello %s!' % name)
    addConfig(USER_PREFIX + 'name', name)

    files = list(range(0, 5))
    # https://click.palletsprojects.com/en/7.x/utils/#showing-progress-bars
    with click.progressbar(files, label='Registering the user....', ) as bar:
        for file in bar:
            time.sleep(random.random())

    # adding default actions during registration, user can change them by using commands later
    for gesture in ctx.gestures:
        key = GESTURE_PREFIX + gesture
        addConfig(key, key)
        click.echo(f"Gesture {gesture} is added by default")
    for action in ctx.actions:
        key = ACTION_PREFIX + action['actionName']
        addConfig(key, action['action'])
        click.echo(f"Action {action['actionName']} is added by default")
    for itr in range(len(ctx.gestures)):
        key = MAP_PREFIX + ctx.gestures[itr]
        addConfig(key, ctx.actions[itr]['actionName'])
        click.echo(f"{ctx.gestures[itr]} mapped to {ctx.actions[itr]['actionName']} by default")

    click.echo(f"User '{name}' registered successfully!")


@cli.command()
@pass_setting
def start(ctx):
    """Starts the application in the background"""
    if(checkIfUserValid(ctx.name)):
        click.echo('Please register first.')
        exit(1)
    click.echo('Starting.....')
    action_name = getConfig(MAP_PREFIX+ctx.gestures[0])
    keyboard_command = getConfig(ACTION_PREFIX+action_name)
    predictor = GesturePrediction(keyboard_command)
    predictor.live_video()
    click.echo('Application started successfully.')

@cli.command()
@pass_setting
def showactions(ctx):
    """Displays different actions configured"""
    if (checkIfUserValid(ctx.name)):
        click.echo('Please register first.')
        exit(1)
    click.echo('Displaying all registered actions for user: ' + ctx.name)
    printSectionConfig(ACTION_PREFIX)


@cli.command()
@click.option('-a', 'actionName', help='Action Name', required=True, type=str)
@click.option('-s', 'shortcut', help='Keyboard shortcut for the action, separated by +', required=True, type=str)
@pass_setting
def addaction(ctx, actionName, shortcut):
    """Add an action with a name and keyboard shortcut which you want to trigger using hand gesture app"""
    if (checkIfUserValid(ctx.name)):
        click.echo('Please register first.')
        exit(1)
    click.echo('Adding action with name:{} and command:{}'.format(actionName, shortcut))
    # TODO: how to handle cross platform keys: Mac / Windows
    key = ACTION_PREFIX + actionName
    addConfig(key, shortcut)
    click.echo('Action %s added.' % actionName)


@cli.command()
@pass_setting
def showgestures(ctx):
    """Displays different gestures configured"""
    if checkIfUserValid(ctx.name):
        click.echo('Please register first.')
        exit(1)
    click.echo('Displaying all registered gestures for user: ' + ctx.name)
    printSectionConfig(GESTURE_PREFIX)


@cli.command()
@click.option('-g', 'gestureName', help='Gesture Name', required=True, type=str)
@pass_setting
def addgesture(ctx, gestureName):
    """Configure a hand gesture in the app"""
    if checkIfUserValid(ctx.name):
        click.echo('Please register first.')
        exit(1)
    click.echo('Adding gesture with name:{}'.format(gestureName))
    # TODO: trigger model train here, and add a corresponding identifier (id) which will be used to identify this gesture in ML model
    key = GESTURE_PREFIX + gestureName
    addConfig(key, key)
    click.echo(f'Gesture {gestureName} added.')


@cli.command()
@click.option('-a', 'actionName', help='Action Name', required=True, type=str)
@click.option('-g', 'gestureName', help='Gesture Name', required=True, type=str)
@pass_setting
def mapActionWithGesture(ctx, actionName, gestureName):
    """Map a configured action with a gesture configured with the app."""
    if checkIfUserValid(ctx.name):
        click.echo('Please register first.')
        exit(1)
    key = MAP_PREFIX + gestureName
    addConfig(key, actionName)
    click.echo('Mapping Complete!')


if __name__ == "__main__":
    cli()
